# flake8: noqa
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hello.proto

import sys


_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="hello.proto",
    package="helloworld",
    syntax="proto3",
    serialized_options=None,
    serialized_pb=_b(
        '\n\x0bhello.proto\x12\nhelloworld"\x1c\n\x0cHelloRequest\x12\x0c\n\x04name\x18\x01 \x01(\t"\x1d\n\nHelloReply\x12\x0f\n\x07message\x18\x01 \x01(\t2\xa2\x02\n\x05Hello\x12>\n\x08SayHello\x12\x18.helloworld.HelloRequest\x1a\x16.helloworld.HelloReply"\x00\x12\x45\n\rSayHelloTwice\x12\x18.helloworld.HelloRequest\x1a\x16.helloworld.HelloReply"\x00\x30\x01\x12L\n\x12SayHelloRepeatedly\x12\x18.helloworld.HelloRequest\x1a\x16.helloworld.HelloReply"\x00(\x01\x30\x01\x12\x44\n\x0cSayHelloLast\x12\x18.helloworld.HelloRequest\x1a\x16.helloworld.HelloReply"\x00(\x01\x62\x06proto3'
    ),
)


_HELLOREQUEST = _descriptor.Descriptor(
    name="HelloRequest",
    full_name="helloworld.HelloRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="name",
            full_name="helloworld.HelloRequest.name",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=27,
    serialized_end=55,
)


_HELLOREPLY = _descriptor.Descriptor(
    name="HelloReply",
    full_name="helloworld.HelloReply",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="message",
            full_name="helloworld.HelloReply.message",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=57,
    serialized_end=86,
)

DESCRIPTOR.message_types_by_name["HelloRequest"] = _HELLOREQUEST
DESCRIPTOR.message_types_by_name["HelloReply"] = _HELLOREPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

HelloRequest = _reflection.GeneratedProtocolMessageType(
    "HelloRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _HELLOREQUEST,
        "__module__": "hello_pb2"
        # @@protoc_insertion_point(class_scope:helloworld.HelloRequest)
    },
)
_sym_db.RegisterMessage(HelloRequest)

HelloReply = _reflection.GeneratedProtocolMessageType(
    "HelloReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _HELLOREPLY,
        "__module__": "hello_pb2"
        # @@protoc_insertion_point(class_scope:helloworld.HelloReply)
    },
)
_sym_db.RegisterMessage(HelloReply)


_HELLO = _descriptor.ServiceDescriptor(
    name="Hello",
    full_name="helloworld.Hello",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    serialized_start=89,
    serialized_end=379,
    methods=[
        _descriptor.MethodDescriptor(
            name="SayHello",
            full_name="helloworld.Hello.SayHello",
            index=0,
            containing_service=None,
            input_type=_HELLOREQUEST,
            output_type=_HELLOREPLY,
            serialized_options=None,
        ),
        _descriptor.MethodDescriptor(
            name="SayHelloTwice",
            full_name="helloworld.Hello.SayHelloTwice",
            index=1,
            containing_service=None,
            input_type=_HELLOREQUEST,
            output_type=_HELLOREPLY,
            serialized_options=None,
        ),
        _descriptor.MethodDescriptor(
            name="SayHelloRepeatedly",
            full_name="helloworld.Hello.SayHelloRepeatedly",
            index=2,
            containing_service=None,
            input_type=_HELLOREQUEST,
            output_type=_HELLOREPLY,
            serialized_options=None,
        ),
        _descriptor.MethodDescriptor(
            name="SayHelloLast",
            full_name="helloworld.Hello.SayHelloLast",
            index=3,
            containing_service=None,
            input_type=_HELLOREQUEST,
            output_type=_HELLOREPLY,
            serialized_options=None,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_HELLO)

DESCRIPTOR.services_by_name["Hello"] = _HELLO

# @@protoc_insertion_point(module_scope)
