# defaults
APP = "redis"
DEFAULT_SERVICE = "redis"

# net extension
DB = "out.redis_db"

# standard tags
RAWCMD = "redis.raw_command"
CMD = "redis.command"
ARGS_LEN = "redis.args_length"
PIPELINE_LEN = "redis.pipeline_length"
PIPELINE_AGE = "redis.pipeline_age"
